#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""grebakker - Tests for continuing backups."""
# =============================================================================
__author__     = "Daniel Krajzewicz"
__copyright__  = "Copyright 2025, Daniel Krajzewicz"
__credits__    = ["Daniel Krajzewicz"]
__license__    = "GPL"
__version__    = "0.4.2"
__maintainer__ = "Daniel Krajzewicz"
__email__      = "daniel@krajzewicz.de"
__status__     = "Development"
# =============================================================================
# - https://github.com/dkrajzew/degrotesque
# - http://www.krajzewicz.de/docs/degrotesque/index.html
# - http://www.krajzewicz.de
# =============================================================================


# --- imports -----------------------------------------------------------------
import sys
import os
sys.path.append(os.path.join(os.path.split(__file__)[0], "..", "grebakker"))
import grebakker
from util import pname, TEST_PATH, check_def, prepare, check_generated, pdirtime


# --- definitions -------------------------------------------------------------
DEFINITION1 = """{
    "destination": "d/",
    "copy": [ 
        "document1.txt",
        "something_else1.csv"
    ],
    "compress": [ 
        { "name": "subfolder" }
    ]
}
"""


# --- test functions ----------------------------------------------------------
def test_main_continue_clean_v0(capsys, tmp_path):
    """Test behaviour if no arguments are given"""
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set1")])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pname(captured.out) == ""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION1)

def test_main_continue_clean_v1(capsys, tmp_path):
    """Test behaviour if no arguments are given"""
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set1"), "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set1'...
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION1)


def test_main_continue_log_exists_plain_v0(capsys, tmp_path):
    """Test behaviour if no arguments are given"""
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path)
    (tmp_path / "grebakker_log.csv").write_text("initial content\n") # generate existing log file
    os.chdir(tmp_path) # !!!
    try:
        grebakker.main(["backup", dstroot, str(tmp_path / "set1")])
        assert False # pragma: no cover
    except SystemExit as e:
        assert type(e)==type(SystemExit())
        assert e.code==2
    captured = capsys.readouterr()
    assert pname(captured.err) == """grebakker: error: a log file exists, but it is not defined whether to restart or to continue.
"""
    assert pname(captured.out) == ""
    #check_generated(tmp_path, actions, "csv")
    #check_def(tmp_path, DEFINITION1)

def test_main_continue_log_exists_plain_v1(capsys, tmp_path):
    """Test behaviour if no arguments are given"""
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path)
    (tmp_path / "grebakker_log.csv").write_text("initial content\n") # generate existing log file
    os.chdir(tmp_path) # !!!
    try:
        grebakker.main(["backup", dstroot, str(tmp_path / "set1"), "-v"])
        assert False # pragma: no cover
    except SystemExit as e:
        assert type(e)==type(SystemExit())
        assert e.code==2
    captured = capsys.readouterr()
    assert pname(captured.err) == """grebakker: error: a log file exists, but it is not defined whether to restart or to continue.
"""
    assert pname(captured.out) == ""
    #check_generated(tmp_path, actions, "csv")
    #check_def(tmp_path, DEFINITION1)


def test_main_continue_log_exists_continue_v0(capsys, tmp_path):
    """Test behaviour if no arguments are given"""
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path)
    (tmp_path / "grebakker_log.csv").write_text("initial content\n") # generate existing log file
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set1"), "--continue"])
    assert ret == 0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pname(captured.out) == """A log file exists; contents will be appended.
"""
    check_generated(tmp_path, actions, dst, "csv", log_head="initial content\n")
    check_def(tmp_path, DEFINITION1)

def test_main_continue_log_exists_continue_v1(capsys, tmp_path):
    """Test behaviour if no arguments are given"""
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path)
    (tmp_path / "grebakker_log.csv").write_text("initial content\n") # generate existing log file
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set1"), "--continue", "-v"])
    assert ret == 0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """A log file exists; contents will be appended.
Starting...
Processing '<DIR>/set1'...
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv", log_head="initial content\n")
    check_def(tmp_path, DEFINITION1)

def test_main_continue_log_exists_continue_csv(capsys, tmp_path):
    """Test behaviour if no arguments are given"""
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path)
    (tmp_path / "grebakker_log.csv").write_text("initial content\n") # generate existing log file
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set1"), "--continue", "--log-format", "csv"])
    assert ret == 0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pname(captured.out) == """A log file exists; contents will be appended.
"""
    check_generated(tmp_path, actions, dst, "csv", log_head="initial content\n")
    check_def(tmp_path, DEFINITION1)

def test_main_continue_log_exists_continue_json(capsys, tmp_path):
    """Test behaviour if no arguments are given"""
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path)
    (tmp_path / "grebakker_log.csv").write_text("initial content\n") # generate existing log file
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set1"), "--continue", "--log-format", "json"])
    assert ret == 0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pname(captured.out) == """A log file exists; contents will be appended.
"""
    check_generated(tmp_path, actions, dst, "json", log_head="initial content\n")
    check_def(tmp_path, DEFINITION1)


def test_main_continue_log_exists_restart_v0(capsys, tmp_path):
    """Test behaviour if no arguments are given"""
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path)
    (tmp_path / "grebakker_log.csv").write_text("initial content\n") # generate existing log file
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set1"), "--log-restart"])
    assert ret == 0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pname(captured.out) == """The existing log file will be replaced by new contents.
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION1)

def test_main_continue_log_exists_restart_v1(capsys, tmp_path):
    """Test behaviour if no arguments are given"""
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path)
    (tmp_path / "grebakker_log.csv").write_text("initial content\n") # generate existing log file
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set1"), "--log-restart", "-v"])
    assert ret == 0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """The existing log file will be replaced by new contents.
Starting...
Processing '<DIR>/set1'...
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""grebakker - Tests for continuing backups."""
# =============================================================================
__author__     = "Daniel Krajzewicz"
__copyright__  = "Copyright 2025, Daniel Krajzewicz"
__credits__    = ["Daniel Krajzewicz"]
__license__    = "GPL"
__version__    = "0.4.2"
__maintainer__ = "Daniel Krajzewicz"
__email__      = "daniel@krajzewicz.de"
__status__     = "Development"
# =============================================================================
# - https://github.com/dkrajzew/degrotesque
# - http://www.krajzewicz.de/docs/degrotesque/index.html
# - http://www.krajzewicz.de
# =============================================================================


# --- imports -----------------------------------------------------------------
import sys
import os
import shutil
sys.path.append(os.path.join(os.path.split(__file__)[0], "..", "grebakker"))
import grebakker
from util import pname, TEST_PATH, check_def, prepare, check_generated, pdirtime


# --- definitions -------------------------------------------------------------
DEFINITION1A = """{
    "destination": "d/",
    "copy": [ 
        "document.txt",
        "something_else.csv"
    ],
    "compress": [ 
        "subfolder2"
    ],
    "subfolders": [ "subfolder1" ]
}
"""
DEFINITION1B = """{
    "destination": "d/subfolder1",
    "compress": [ 
        "subfolder11",
        "subfolder12"
    ]
}
"""

DEFINITION2A = """{
    "destination": "d/",
    "copy": [ 
        "document.txt",
        "something_else.csv"
    ],
    "compress": [ 
        "subfolder2"
    ],
    "subfolders": [ "subfolder1" ]
}
"""
DEFINITION2B = """{
    "destination": "d/subfolder1",
    "compress": [ 
        { "name": "subfolder11", "exclude": ["*.csv"] },
        { "name": "subfolder12", "exclude": ["*.txt"] }
    ]
}
"""

DEFINITION3 = """{
    "destination": "d/",
    "copy": [ 
        "document.txt",
        "something_else.csv"
    ],
    "compress": [ 
        { "name": "subfolder1", "exclude" : [ "subfolder12" ] },
        "subfolder2"
    ]
}
"""

DEFINITION4 = """{
    "destination": "d/",
    "copy": [ 
        { "name": "subfolder1", "exclude" : [ "subfolder12" ] }
    ]
}
"""

# --- test functions ----------------------------------------------------------
def test_main_plain_v0(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1A, "set2", tmp_path, add_defs={"subfolder1": DEFINITION1B})
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2")])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pname(captured.out) == ""
    #print(actions)
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION1A)

def test_main_plain_v1(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1A, "set2", tmp_path, add_defs={"subfolder1": DEFINITION1B})
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Processing '<DIR>/set2/subfolder1'...
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION1A)

def test_main_plain_v2(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1A, "set2", tmp_path, add_defs={"subfolder1": DEFINITION1B})
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder2'... done. (<DUR>)
 Processing '<DIR>/set2/subfolder1'...
  Compressing '<DIR>/set2/subfolder1/subfolder11'... done. (<DUR>)
  Compressing '<DIR>/set2/subfolder1/subfolder12'... done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION1A)


def test_main_exclude_compress_files_in_folder1_v2(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION2A, "set2", tmp_path, add_defs={"subfolder1": DEFINITION2B})
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder2'... done. (<DUR>)
 Processing '<DIR>/set2/subfolder1'...
  Compressing '<DIR>/set2/subfolder1/subfolder11'... done. (<DUR>)
  Compressing '<DIR>/set2/subfolder1/subfolder12'... done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv", 
        testfiles={"subfolder11": "subfolder11_nocsv.zip", "subfolder12": "subfolder12_notxt.zip"})
    check_def(tmp_path, DEFINITION2A)


def test_main_exclude_compress_folder_in_folder1_v2(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION3, "set2", tmp_path)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder1'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder2'... done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv", testfiles={"subfolder1": "subfolder1_nosubfolder12.zip"})
    check_def(tmp_path, DEFINITION3)


def test_main_exclude_copy_folder_in_folder1_v2(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION4, "set2", tmp_path)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/subfolder1'... done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION4)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""grebakker - greyrat's backupper for hackers."""
# ===========================================================================
__author__     = "Daniel Krajzewicz"
__copyright__  = "Copyright 2025, Daniel Krajzewicz"
__credits__    = "Daniel Krajzewicz"
__license__    = "GPL"
__version__    = "0.4.2"
__maintainer__ = "Daniel Krajzewicz"
__email__      = "daniel@krajzewicz.de"
__status__     = "Development"
# ===========================================================================
# - https://github.com/dkrajzew/grebakker
# - http://www.krajzewicz.de
# ===========================================================================


# --- imports ---------------------------------------------------------------
import json
import sys
import os
import datetime
import argparse
import configparser
import shutil
import zipfile
import fnmatch
from typing import List, Dict, Any, Union, Generator


# --- class definitions -----------------------------------------------------
class Log:
    """Handles logging of performed actions in CSV or JSON format."""

    def __init__(self, name: str, restart: bool, log_format: str, off: bool) -> None:
        """Initialize the Log object.

        Args:
            name (str): Path to the log file
            restart (bool): If True, overwrites the existing log file
            log_format (str): Format of the log ('csv' or 'json')
            off (bool): If True, disables logging
        """
        self._format = log_format
        self._written = 0
        self._output = None
        self._name = name
        if off:
            return
        mode = "w" if restart else "a"
        self._output = open(name, mode, encoding="utf-8")
        if self._format=="json":
            self._output.write('[\n')

    def write(self, act: str, src: str, dst: str, duration: str) -> None:
        """Write a log entry about a performed action.

        Args:
            act (str): Action performed
            src (str): Source path
            dst (str): Destination path
            duration (str): Duration of the action
        """
        if self._output is None:
            return
        if self._format=="csv":
            self._output.write(f'"{act}";"{src}";"{dst}";"{duration}"\n')
        elif self._format=="json":
            if self._written!=0:
                self._output.write(',\n')
            self._output.write('    {"action": "' + act + '", "src": "' + src + '", "dst": "' + dst + '", "duration": "' + duration + '"}')
        self._output.flush()
        self._written += 1

    def error(self, error: str) -> None:
        """Write a log entry about an error.

        Args:
            error (str): The error occured
        """
        if self._output is None:
            return
        if self._format=="csv":
            self._output.write(f'"{error}"\n')
        elif self._format=="json":
            if self._written!=0:
                self._output.write(',\n')
            self._output.write('    {"error": "' + error + '"}')
        self._output.flush()
        self._written += 1

    def close(self) -> None:
        """Close the log file."""
        if self._output is None:
            return
        if self._format=="json":
            self._output.write('\n]\n')
        self._output.close()


class Grebakker:
    """Perform backup operations."""

    def __init__(self, dest: str, log: Log, verbosity: int):
        """Initialize the Grebakker object.

        Args:
            dest (str): Destination directory for backups
            log (Log): Log object for recording actions
            verbosity (int): Verbosity level for output
        """
        self._dest = dest
        self._log = log
        self._verbosity = verbosity
        self._line_ended = True

    def _action_begin(self, mml_action: str, path: str, level: int) -> datetime.datetime:
        """Report the beginning of an action, return starting time.

        Args:
            mml_action (str): Description of the action
            path (str): Path involved in the action
            level (int): Indentation level

        Returns:
            (datetime.datetime): Start time of the action.
        """
        if self._verbosity>1:
            print(f"{self._i(level+1)}{mml_action} '{path}'... ", end="", flush=True)
            self._line_ended = False
        return datetime.datetime.now()

    def _action_end(self, action: str, path: str, dst: str, level: int, t1: datetime.datetime) -> None:
        """Report the end of an action and logs it.

        Args:
            action (str): Action performed
            path (str): Source path
            dst (str): Destination path
            level (int): Indentation level
            t1 (datetime.datetime): Start time of the action
        """
        t2 = datetime.datetime.now()
        self._log.write(action, path, dst, str(t2-t1))
        if self._verbosity>1:
            print(f"done. ({t2-t1})")
            self._line_ended = True

    def _yield_files(self, src: str, exclude: List[str]) -> Generator:
        """Yield files from the source directory, excluding specified patterns.

        Args:
            src (str): Source directory.
            exclude (List[str]): List of patterns to exclude.

        Yields:
            source_path (str): Relative path to each file
        """
        for root, dirs, files in os.walk(src):
            dirs[:] = [d for d in dirs if d not in exclude]
            for file in files:
                srcf = os.path.relpath(os.path.join(root, file), src)
                use = True
                for e in exclude:
                    if fnmatch.fnmatchcase(srcf, e):
                        use = False
                        break
                if not use:
                    continue
                yield os.path.relpath(os.path.join(root, file), os.path.join(src, ".."))

    def _i(self, level: int) -> str:
        """Return indentation spaces for the given output level.

        Args:
            level (int): Indentation level.

        Returns:
            (str): Indentation spaces
        """
        return ' '*level

    def _destination_exists(self, action: str, src: str, dst: str, level: int) -> bool:
        """Determine the destination path for an action.

        Args:
            action (str): Action being performed
            src (str): Source path
            dst (str): Destination path
            level (int): Indentation level

        Returns:
            (str or None): Destination path or None if the destination file exists.
        """
        if not os.path.exists(dst):
            return False
        self._log.write(action, src, dst, "skipped")
        if self._verbosity>1:
            if not self._line_ended:
                print()
                self._line_ended = True
            print(f"{self._i(level+1)}Skipping '{dst}' - exists.")
        return True


    def copy(self, src_root: str, item: Union[str, Dict[str, str]], dst_root: str, level: int) -> None:
        """Copy files or directories from source to destination.

        Args:
            src_root (str): Root source directory
            item (Union[str, Dict[str, str]]): Item to copy. If dict, may contain 'name' and 'exclude'
            dst_root (str): Root destination directory
            level (int): Reporting level

        Raises:
            FileNotFoundError: If the source path does not exist.
        """
        path = item if isinstance(item, str) else item["name"]
        src = os.path.join(src_root, path)
        if not os.path.exists(src):
            self._log.error(f"file/folder '{src}' to copy does not exist")
            raise FileNotFoundError(f"file/folder '{src}' to copy does not exist")
        dst = os.path.join(dst_root, path)
        if os.path.isfile(src):
            if self._destination_exists("copy", src, dst, level):
                return
            t1 = self._action_begin("Copying", src, level)
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            shutil.copy2(src, dst)
            self._action_end("copy", src, dst, level, t1)
            return
        exclude = [] if "exclude" not in item else item["exclude"]
        exclude = [exclude] if isinstance(exclude, str) else exclude
        t1 = self._action_begin("Copying", src, level)
        for file in self._yield_files(src, list(exclude)):
            fsrc = os.path.abspath(os.path.join(src, "..", file))
            fdst = os.path.abspath(os.path.join(dst, "..", file))
            if self._destination_exists("copy", os.path.join(src, file), fdst, level):
                continue
            os.makedirs(os.path.dirname(fdst), exist_ok=True)
            shutil.copy2(fsrc, fdst)
        self._action_end("copy", src, dst, level, t1)


    def compress(self, root: str, item: Union[str, Dict[str, str]], dst_root: str, level: int) -> None:
        """Compress files or directories into a ZIP archive.

        Args:
            root (str): Root source directory
            item (Union[str, Dict[str, str]]): Item to compress. If dict, may contain 'name' and 'exclude'
            dst_root (str): Root destination directory
            level (int): Reporting level

        Raises:
            FileNotFoundError: If the source path does not exist
        """
        path = item if isinstance(item, str) else item["name"]
        src = os.path.join(root, path)
        if not os.path.exists(src):
            self._log.error(f"file/folder '{src}' to compress does not exist")
            raise FileNotFoundError(f"file/folder '{src}' to compress does not exist")
        dst = os.path.join(dst_root, path) + ".zip"
        if self._destination_exists("compress", src, dst, level):
            return
        exclude = [] if "exclude" not in item else item["exclude"]
        exclude = [exclude] if isinstance(exclude, str) else exclude
        t1 = self._action_begin("Compressing", src, level)
        zipf = zipfile.ZipFile(dst, "w", zipfile.ZIP_DEFLATED, compresslevel=9)
        if os.path.isfile(src):
            dstf = os.path.relpath(src, os.path.join(src, '..'))
            zipf.write(src, dstf)
        else:
            for file in self._yield_files(src, list(exclude)):
                fsrc = os.path.join(src, "..", file)
                dstf = os.path.relpath(os.path.join(root, file), os.path.join(src, '..'))
                zipf.write(fsrc, dstf)
        zipf.close()
        self._action_end("compress", src, dst, level, t1)


    def decompress(self, root: str, item: Union[str, Dict[str, str]], dst_root: str, level: int) -> None:
        """Decompress ZIP archives.

        Args:
            root (str): Root source directory
            item (Union[str, Dict[str, str]]): Item to compress. If dict, may contain 'name' and 'exclude'
            dst_root (str): Root destination directory
            level (int): Reporting level

        Raises:
            FileNotFoundError: If the source path does not exist
        """
        path = item if isinstance(item, str) else item["name"]
        src = os.path.join(root, path) + ".zip"
        if not os.path.exists(src):
            self._log.error(f"file/folder '{src}' to decompress does not exist")
            raise FileNotFoundError(f"file/folder '{src}' to decompress does not exist")
        dst = dst_root#os.path.join(dst_root, path)
        #if self._destination_exists("decompress", src, dst, level):
        #    return
        t1 = self._action_begin("Decompressing", src, level)
        with zipfile.ZipFile(src, 'r') as zip_ref:
            zip_ref.extractall(dst)
        self._action_end("decompress", src, dst, level, t1)


    def run(self, action: str, root: str, level: int=0) -> None:
        """Perform an action.

        Args:
            action (str): The action to perform
            root (str): Root source directory
            level (int): Reporting level
        """
        if action not in ["backup", "restore"]:
            raise ValueError(f"unkown action '{action}'")
        # init
        if self._verbosity>0:
            print(f"{self._i(level)}Processing '{root}'...")
            self._line_ended = True
        definition = None
        with open(os.path.join(root, "grebakker.json"), encoding="utf-8") as fd:
            definition = json.load(fd)
        dst_path = os.path.join(self._dest, definition["destination"])
        os.makedirs(dst_path, exist_ok=True)
        # copy
        for path in [] if "copy" not in definition else definition["copy"]:
            self.copy(root, path, dst_path, level)
        # compress
        for path in [] if "compress" not in definition else definition["compress"]:
            if action=="backup":
                self.compress(root, path, dst_path, level)
            elif action=="restore":
                self.decompress(root, path, dst_path, level)
        # subfolders
        for sub in [] if "subfolders" not in definition else definition["subfolders"]:
            path = sub if type(sub)==str else sub["name"]
            src = os.path.join(root, path)
            if not os.path.exists(src):
                self._log.error(f"file/folder '{src}' to recurse into does not exist")
                raise FileNotFoundError(f"file/folder '{src}' to recurse into does not exist")
            dst = os.path.join(dst_path, path)
            # !!! add an option for what to do
            #if self._destination_exists("sub", src, dst, level):
            #    pass
            self._log.write("sub", src, dst, "0:00:00")
            self.run(action, os.path.join(root, sub), level+1)
        shutil.copy2(os.path.join(root, "grebakker.json"), dst_path)
        if level==0:
            self._log.close()
            if self._log._written!=0:
                shutil.move(self._log._name, os.path.join(dst_path, self._log._name))



# --- functions -------------------------------------------------------------
def main(arguments: List[str] = []) -> int:
    """Run grebaker using the given parameters.

    Args:
        arguments (List[str]): A list of command line arguments

    Returns:
        (int): The exit code (0 for success)
    """
    # parse options
    # https://stackoverflow.com/questions/3609852/which-is-the-best-way-to-allow-configuration-options-be-overridden-at-the-comman
    defaults = {}
    conf_parser = argparse.ArgumentParser(prog='grebakker', add_help=False)
    conf_parser.add_argument("-c", "--config", metavar="FILE", help="Reads the named configuration file")
    args, remaining_argv = conf_parser.parse_known_args(arguments)
    if args.config is not None:
        if not os.path.exists(args.config):
            print ("grebakker: error: configuration file '%s' does not exist" % str(args.config), file=sys.stderr)
            raise SystemExit(2)
        config = configparser.ConfigParser()
        config.read([args.config])
        defaults.update(dict(config.items("grebakker")))
    parser = argparse.ArgumentParser(prog='grebakker', parents=[conf_parser],
                                     description="greyrat's backupper for hackers",
                                     epilog='(c) Daniel Krajzewicz 2025')
    parser.add_argument("action" if "action" not in defaults else "--action")
    parser.add_argument("destination" if "destination" not in defaults else "--destination")
    parser.add_argument("definition" if "definition" not in defaults else "--definition")
    parser.add_argument('--version', action='version', version='%(prog)s 0.4.2')
    parser.add_argument('--continue', dest="cont", action="store_true", help="Continues a stopped backup.")
    parser.add_argument('--log-name', default="grebakker_log.csv", help="Change logfile name (default: 'grebakker_log.csv').")
    parser.add_argument('--log-restart', action="store_true", help="An existing logfile will be removed.")
    parser.add_argument('--log-off', action="store_true", help="Does not generate a log file.")
    parser.add_argument('--log-format', default="csv", choices=['csv', 'json'], help="Select log format to use ['csv', 'json']")
    """
    parser.add_argument('--compress-max-size', default="2g", help="Defines maximum archive size (default: 2g).")
    parser.add_argument('--compress-max-size-abort', action="store_true", help="Stop when an archive is bigger than the maximum size.")
    parser.add_argument('--compress-max-threads', type=int, default=1, help="Defines maximum thread number.")
    """
    #parser.add_argument('-e', '--excluded-log', metavar="FILE", default=None, help="Writes excluded files and folders to FILE.")
    parser.add_argument('-v', '--verbose', action='count', default=0, help="Increases verbosity level (up to 2).") # https://stackoverflow.com/questions/6076690/verbose-level-with-argparse-and-multiple-v-options
    parser.set_defaults(**defaults)
    args = parser.parse_args(remaining_argv)
    verbosity = int(args.verbose)
    # check
    errors = []
    if args.action not in ["backup", "restore"]:
        errors.append(f"unkown action '{args.action}'")
    if len(errors)!=0:
        for e in errors:
            print(f"grebakker: error: {e}", file=sys.stderr)
        raise SystemExit(2)
    #
    if os.path.exists(args.log_name):
        if not args.cont and not args.log_restart:
            print("grebakker: error: a log file exists, but it is not defined whether to restart or to continue.", file=sys.stderr)
            raise SystemExit(2)
        if args.cont:
            print("A log file exists; contents will be appended.")
        if args.log_restart:
            print("The existing log file will be replaced by new contents.")
    if not os.path.exists(args.destination):
        os.makedirs(args.destination, exist_ok=True)
    #
    ret = 0
    if verbosity>0:
        print("Starting...")
    t1 = datetime.datetime.now()
    log = Log(args.log_name, args.log_restart, args.log_format, args.log_off)
    grebakker = Grebakker(args.destination, log, verbosity)
    try:
        grebakker.run(args.action, args.definition)
    except FileNotFoundError as e:
        print(f"grebakker: error: {e}.", file=sys.stderr)
        ret = 2
    t2 = datetime.datetime.now()
    if verbosity>0 and ret==0:
        print(f"Completed after {(t2-t1)}")
    return ret


def script_run() -> int:
    """Execute from command line."""
    sys.exit(main(sys.argv[1:])) # pragma: no cover


# -- main check
if __name__ == '__main__':
    sys.exit(main(sys.argv[1:])) # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""grebakker - Tests for the main method."""
# =============================================================================
__author__     = "Daniel Krajzewicz"
__copyright__  = "Copyright 2025, Daniel Krajzewicz"
__credits__    = ["Daniel Krajzewicz"]
__license__    = "GPL"
__version__    = "0.4.2"
__maintainer__ = "Daniel Krajzewicz"
__email__      = "daniel@krajzewicz.de"
__status__     = "Development"
# =============================================================================
# - https://github.com/dkrajzew/degrotesque
# - http://www.krajzewicz.de/docs/degrotesque/index.html
# - http://www.krajzewicz.de
# =============================================================================


# --- imports -----------------------------------------------------------------
import sys
import os
sys.path.append(os.path.join(os.path.split(__file__)[0], "..", "grebakker"))
import grebakker



# --- helper functions --------------------------------------------------------
def patch(string):
    return string.replace("__main__.py", "degrotesque").replace("pytest", "degrotesque").replace("optional arguments", "options")



# --- test functions ----------------------------------------------------------
def test_main_empty1(capsys):
    """Test behaviour if no arguments are given"""
    try:
        ret = grebakker.main([])
        assert False # pragma: no cover
    except SystemExit as e:
        assert type(e)==type(SystemExit())
        assert e.code==2
    captured = capsys.readouterr()
    assert patch(captured.err) == """usage: grebakker [-h] [-c FILE] [--version] [--continue] [--log-name LOG_NAME]
                 [--log-restart] [--log-off] [--log-format {csv,json}] [-v]
                 action destination definition
grebakker: error: the following arguments are required: action, destination, definition
"""
    assert patch(captured.out) == ""


def test_main_empty2(capsys):
    """Test behaviour if no arguments are given"""
    try:
        ret = grebakker.main()
        assert False # pragma: no cover
    except SystemExit as e:
        assert type(e)==type(SystemExit())
        assert e.code==2
    captured = capsys.readouterr()
    assert patch(captured.err) == """usage: grebakker [-h] [-c FILE] [--version] [--continue] [--log-name LOG_NAME]
                 [--log-restart] [--log-off] [--log-format {csv,json}] [-v]
                 action destination definition
grebakker: error: the following arguments are required: action, destination, definition
"""
    assert patch(captured.out) == ""


def test_main_help(capsys):
    """Test behaviour when help is wished"""
    try:
        grebakker.main(["--help"])
        assert False # pragma: no cover
    except SystemExit as e:
        assert type(e)==type(SystemExit())
        assert e.code==0
    captured = capsys.readouterr()
    assert patch(captured.out) == """usage: grebakker [-h] [-c FILE] [--version] [--continue] [--log-name LOG_NAME]
                 [--log-restart] [--log-off] [--log-format {csv,json}] [-v]
                 action destination definition

greyrat's backupper for hackers

positional arguments:
  action
  destination
  definition

options:
  -h, --help            show this help message and exit
  -c FILE, --config FILE
                        Reads the named configuration file
  --version             show program's version number and exit
  --continue            Continues a stopped backup.
  --log-name LOG_NAME   Change logfile name (default: 'grebakker_log.csv').
  --log-restart         An existing logfile will be removed.
  --log-off             Does not generate a log file.
  --log-format {csv,json}
                        Select log format to use ['csv', 'json']
  -v, --verbose         Increases verbosity level (up to 2).

(c) Daniel Krajzewicz 2025
"""
    assert captured.err == ""


def test_main_version(capsys):
    """Test behaviour when version information is wished"""
    try:
        grebakker.main(["--version"])
        assert False # pragma: no cover
    except SystemExit as e:
        assert type(e)==type(SystemExit())
        assert e.code==0
    captured = capsys.readouterr()
    assert patch(captured.out) == """grebakker 0.4.2
"""
    assert patch(captured.err) == ""




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""grebakker - Tests for continuing backups."""
# =============================================================================
__author__     = "Daniel Krajzewicz"
__copyright__  = "Copyright 2025, Daniel Krajzewicz"
__credits__    = ["Daniel Krajzewicz"]
__license__    = "GPL"
__version__    = "0.4.2"
__maintainer__ = "Daniel Krajzewicz"
__email__      = "daniel@krajzewicz.de"
__status__     = "Development"
# =============================================================================
# - https://github.com/dkrajzew/degrotesque
# - http://www.krajzewicz.de/docs/degrotesque/index.html
# - http://www.krajzewicz.de
# =============================================================================


# --- imports -----------------------------------------------------------------
import sys
import os
import shutil
sys.path.append(os.path.join(os.path.split(__file__)[0], "..", "grebakker"))
import grebakker
from util import pname, TEST_PATH, check_def, prepare, check_generated, pdirtime


# --- definitions -------------------------------------------------------------
DEFINITION1 = """{
    "destination": "d/",
    "copy": [ 
        "document1.txt",
        "something_else1.csv"
    ],
    "compress": [ 
        { "name": "subfolder" }
    ]
}
"""

DEFINITION2 = """{
    "destination": "d/",
    "copy": [ 
        "subfolder2"
    ]
}
"""

DEFINITION3A = """{
    "destination": "d/",
    "subfolders": [ "subfolder1" ]
}
"""
DEFINITION3B = """{
    "destination": "d/subfolder1",
    "compress": [ 
        { "name": "subfolder11" },
        { "name": "subfolder12" }
    ]
}
"""

# --- test functions ----------------------------------------------------------
def test_main_skip_clean_v2(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set1"), "-v", "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set1'...
 Copying '<DIR>/set1/document1.txt'... done. (<DUR>)
 Copying '<DIR>/set1/something_else1.csv'... done. (<DUR>)
 Compressing '<DIR>/set1/subfolder'... done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION1)


def test_main_skip_doc1_exists_v2(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path, skipped=["document1.txt"])
    os.makedirs(os.path.join(dst, "d"))
    shutil.copy(tmp_path / "set1" / "document1.txt", os.path.join(dstroot, "d", "document1.txt"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set1"), "-v", "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set1'...
 Skipping '<DIR>/backup/d/document1.txt' - exists.
 Copying '<DIR>/set1/something_else1.csv'... done. (<DUR>)
 Compressing '<DIR>/set1/subfolder'... done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION1)


def test_main_skip_archive1_exists_v2(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1, "set1", tmp_path, skipped=["subfolder"])
    os.makedirs(os.path.join(dstroot, "d"))
    shutil.copy(os.path.join(TEST_PATH, "subfolder.zip"), os.path.join(dstroot, "d", "subfolder.zip"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set1"), "-v", "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set1'...
 Copying '<DIR>/set1/document1.txt'... done. (<DUR>)
 Copying '<DIR>/set1/something_else1.csv'... done. (<DUR>)
 Skipping '<DIR>/backup/d/subfolder.zip' - exists.
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION1)


def test_main_skip_folderfile_exists__plain_v2(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION2, "set2", tmp_path)#, skipped=["subfolder"])
    os.makedirs(os.path.join(dstroot, "d"))
    #shutil.copy(os.path.join(TEST_PATH, "subfolder.zip"), os.path.join(dstroot, "d", "subfolder.zip"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/subfolder2'... done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION2)


def test_main_skip_folderfile_exists__exists_v2(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION2, "set2", tmp_path, skipped=["document2.txt"])
    #shutil.copy(os.path.join(TEST_PATH, "subfolder.zip"), os.path.join(dstroot, "d", "subfolder.zip"))
    os.makedirs(os.path.join(dstroot, "d", "subfolder2"))
    shutil.copy(os.path.join(TEST_PATH, "set2", "subfolder2", "document2.txt"), os.path.join(dstroot, "d", "subfolder2", "document2.txt"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/subfolder2'... 
 Skipping '<DIR>/backup/d/subfolder2/document2.txt' - exists.
done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv", log_head='"copy";"<DIR>/set2/subfolder2/subfolder2/document2.txt";"<DIR>/backup/d/subfolder2/document2.txt";"skipped"\n')
    check_def(tmp_path, DEFINITION2)


def test_main_skip_subfolder_exists__plain_v2(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION3A, "set2", tmp_path, add_defs={"subfolder1": DEFINITION3B})
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Processing '<DIR>/set2/subfolder1'...
  Compressing '<DIR>/set2/subfolder1/subfolder11'... done. (<DUR>)
  Compressing '<DIR>/set2/subfolder1/subfolder12'... done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION3A)
    #check_def(tmp_path / "subfolder1", DEFINITION3B)


def test_main_skip_subfolder_exists__exists_v2(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION3A, "set2", tmp_path, add_defs={"subfolder1": DEFINITION3B})
    os.makedirs(os.path.join(dstroot, "d", "subfolder1"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Processing '<DIR>/set2/subfolder1'...
  Compressing '<DIR>/set2/subfolder1/subfolder11'... done. (<DUR>)
  Compressing '<DIR>/set2/subfolder1/subfolder12'... done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION3A)
    #check_def(tmp_path / "subfolder1", DEFINITION3B)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""grebakker - Setup module."""
# ===========================================================================
__author__     = "Daniel Krajzewicz"
__copyright__  = "Copyright 2025, Daniel Krajzewicz"
__credits__    = ["Daniel Krajzewicz"]
__license__    = "GPL"
__version__    = "0.4.2"
__maintainer__ = "Daniel Krajzewicz"
__email__      = "daniel@krajzewicz.de"
__status__     = "Development"
# ===========================================================================
# - https://github.com/dkrajzew/grebakker
# - http://www.krajzewicz.de/docs/grebakker/index.html
# - http://www.krajzewicz.de
# ===========================================================================


# --- imports ---------------------------------------------------------------
import setuptools


# --- definitions -----------------------------------------------------------
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="grebakker",
    version="0.4.2",
    author="dkrajzew",
    author_email="d.krajzewicz@gmail.com",
    description="greyrat's backupper for hackers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://grebakker.readthedocs.org/',
    download_url='http://pypi.python.org/pypi/grebakker',
    project_urls={
        'Documentation': 'https://grebakker.readthedocs.io/',
        'Source': 'https://github.com/dkrajzew/grebakker',
        'Tracker': 'https://github.com/dkrajzew/grebakker/issues',
        'Discussions': 'https://github.com/dkrajzew/grebakker/discussions',
    },
    license='GPL',
    packages = [""],
    package_dir = { "": "grebakker" },
    entry_points = {
        'console_scripts': [
            'grebakker = grebakker:script_run'
        ]
    },
    # see https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Other Audience"
    ],
    python_requires='>=3, <4',
)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""grebakker - Tests for error handling in the main method."""
# =============================================================================
__author__     = "Daniel Krajzewicz"
__copyright__  = "Copyright 2025, Daniel Krajzewicz"
__credits__    = ["Daniel Krajzewicz"]
__license__    = "GPL"
__version__    = "0.4.2"
__maintainer__ = "Daniel Krajzewicz"
__email__      = "daniel@krajzewicz.de"
__status__     = "Development"
# =============================================================================
# - https://github.com/dkrajzew/degrotesque
# - http://www.krajzewicz.de/docs/degrotesque/index.html
# - http://www.krajzewicz.de
# =============================================================================


# --- imports -----------------------------------------------------------------
import sys
import os
import shutil
sys.path.append(os.path.join(os.path.split(__file__)[0], "..", "grebakker"))
import grebakker
from util import pname, TEST_PATH, check_def, prepare, check_generated, pdirtime


DEFINITION1 = """{
    "destination": "d/",
    "copy": [ 
        "document.txt",
        "something_else.csv",
        "subfolder1",
        "subfolder2"
    ]
}
"""

DEFINITION2 = """{
    "destination": "d/",
    "compress": [ 
        "document.txt",
        "subfolder1",
        "subfolder2"
    ]
}
"""

DEFINITION3A = """{
    "destination": "d/",
    "copy": [ 
        "document.txt",
        "something_else.csv"
    ],
    "compress": [ 
        "subfolder2"
    ],
    "subfolders": [ "subfolder1" ]
}
"""
DEFINITION3B = """{
    "destination": "d/subfolder1",
    "compress": [ 
        "subfolder11",
        "subfolder12"
    ]
}
"""


# --- helper functions --------------------------------------------------------
def patch(string):
    return string.replace("__main__.py", "degrotesque").replace("pytest", "degrotesque").replace("optional arguments", "options").replace("choose from 'csv', 'json'", "choose from csv, json")



# --- test functions ----------------------------------------------------------
def test_main_missing3(capsys):
    try:
        ret = grebakker.main([])
        assert False # pragma: no cover
    except SystemExit as e:
        assert type(e)==type(SystemExit())
        assert e.code==2
    captured = capsys.readouterr()
    assert patch(captured.err) == """usage: grebakker [-h] [-c FILE] [--version] [--continue] [--log-name LOG_NAME]
                 [--log-restart] [--log-off] [--log-format {csv,json}] [-v]
                 action destination definition
grebakker: error: the following arguments are required: action, destination, definition
"""
    assert patch(captured.out) == ""

def test_main_missing2(capsys):
    try:
        ret = grebakker.main(["foo"])
        assert False # pragma: no cover
    except SystemExit as e:
        assert type(e)==type(SystemExit())
        assert e.code==2
    captured = capsys.readouterr()
    assert patch(captured.err) == """usage: grebakker [-h] [-c FILE] [--version] [--continue] [--log-name LOG_NAME]
                 [--log-restart] [--log-off] [--log-format {csv,json}] [-v]
                 action destination definition
grebakker: error: the following arguments are required: destination, definition
"""
    assert patch(captured.out) == ""

def test_main_missing1(capsys):
    try:
        ret = grebakker.main(["foo", "bar"])
        assert False # pragma: no cover
    except SystemExit as e:
        assert type(e)==type(SystemExit())
        assert e.code==2
    captured = capsys.readouterr()
    assert patch(captured.err) == """usage: grebakker [-h] [-c FILE] [--version] [--continue] [--log-name LOG_NAME]
                 [--log-restart] [--log-off] [--log-format {csv,json}] [-v]
                 action destination definition
grebakker: error: the following arguments are required: definition
"""
    assert patch(captured.out) == ""


def test_main_unknown_action(capsys):
    try:
        ret = grebakker.main(["foo", "bar", "boo"])
        assert False # pragma: no cover
    except SystemExit as e:
        assert type(e)==type(SystemExit())
        assert e.code==2
    captured = capsys.readouterr()
    assert patch(captured.err) == """grebakker: error: unkown action 'foo'
"""
    assert patch(captured.out) == ""


def test_main_unknown_logformat(capsys):
    try:
        ret = grebakker.main(["backup", "bar", "boo", "--log-format", "foo"])
        assert False # pragma: no cover
    except SystemExit as e:
        assert type(e)==type(SystemExit())
        assert e.code==2
    captured = capsys.readouterr()
    assert patch(captured.err) == """usage: grebakker [-h] [-c FILE] [--version] [--continue] [--log-name LOG_NAME]
                 [--log-restart] [--log-off] [--log-format {csv,json}] [-v]
                 action destination definition
grebakker: error: argument --log-format: invalid choice: 'foo' (choose from csv, json)
"""
    assert patch(captured.out) == ""    


def test_main_copy_missing_valid(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1, "set2", tmp_path)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Copying '<DIR>/set2/subfolder1'... done. (<DUR>)
 Copying '<DIR>/set2/subfolder2'... done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION1)


def test_main_copy_missing_file_default(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1, "set2", tmp_path)
    os.remove(os.path.join(tmp_path, "set2", "something_else.csv"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/something_else.csv' to copy does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION1A)

def test_main_copy_missing_file_none(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1, "set2", tmp_path)
    os.remove(os.path.join(tmp_path, "set2", "something_else.csv"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv", "--log-off"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/something_else.csv' to copy does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION1A)

def test_main_copy_missing_file_csv(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1, "set2", tmp_path)
    os.remove(os.path.join(tmp_path, "set2", "something_else.csv"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv", "--log-format", "csv"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/something_else.csv' to copy does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION1A)

def test_main_copy_missing_file_json(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1, "set2", tmp_path)
    os.remove(os.path.join(tmp_path, "set2", "something_else.csv"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv", "--log-format", "json"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/something_else.csv' to copy does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION1A)


def test_main_copy_missing_folder_default(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1, "set2", tmp_path)
    shutil.rmtree(os.path.join(tmp_path, "set2", "subfolder2"), ignore_errors=True)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/subfolder2' to copy does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Copying '<DIR>/set2/subfolder1'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION1A)

def test_main_copy_missing_folder_none(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1, "set2", tmp_path)
    shutil.rmtree(os.path.join(tmp_path, "set2", "subfolder2"), ignore_errors=True)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv", "--log-off"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/subfolder2' to copy does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Copying '<DIR>/set2/subfolder1'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION1A)

def test_main_copy_missing_folder_csv(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1, "set2", tmp_path)
    shutil.rmtree(os.path.join(tmp_path, "set2", "subfolder2"), ignore_errors=True)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv", "--log-format", "csv"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/subfolder2' to copy does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Copying '<DIR>/set2/subfolder1'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION1A)

def test_main_copy_missing_folder_json(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION1, "set2", tmp_path)
    shutil.rmtree(os.path.join(tmp_path, "set2", "subfolder2"), ignore_errors=True)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv", "--log-format", "json"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/subfolder2' to copy does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Copying '<DIR>/set2/subfolder1'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION1A)


def test_main_compress_missing_valid(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION2, "set2", tmp_path)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Compressing '<DIR>/set2/document.txt'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder1'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder2'... done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION2)

def test_main_compress_missing_file_default(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION2, "set2", tmp_path)
    os.remove(os.path.join(tmp_path, "set2", "document.txt"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/document.txt' to compress does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION2)

def test_main_compress_missing_file_none(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION2, "set2", tmp_path)
    os.remove(os.path.join(tmp_path, "set2", "document.txt"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv", "--log-off"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/document.txt' to compress does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION2)

def test_main_compress_missing_file_csv(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION2, "set2", tmp_path)
    os.remove(os.path.join(tmp_path, "set2", "document.txt"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv", "--log-format", "csv"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/document.txt' to compress does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION2)

def test_main_compress_missing_file_json(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION2, "set2", tmp_path)
    os.remove(os.path.join(tmp_path, "set2", "document.txt"))
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv", "--log-format", "json"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/document.txt' to compress does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION2)


def test_main_compress_missing_folder_default(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION2, "set2", tmp_path)
    shutil.rmtree(os.path.join(tmp_path, "set2", "subfolder2"), ignore_errors=True)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/subfolder2' to compress does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Compressing '<DIR>/set2/document.txt'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder1'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION2)

def test_main_compress_missing_folder_none(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION2, "set2", tmp_path)
    shutil.rmtree(os.path.join(tmp_path, "set2", "subfolder2"), ignore_errors=True)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv", "--log-off"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/subfolder2' to compress does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Compressing '<DIR>/set2/document.txt'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder1'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION2)

def test_main_compress_missing_folder_csv(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION2, "set2", tmp_path)
    shutil.rmtree(os.path.join(tmp_path, "set2", "subfolder2"), ignore_errors=True)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv", "--log-format", "csv"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/subfolder2' to compress does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Compressing '<DIR>/set2/document.txt'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder1'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION2)

def test_main_compress_missing_folder_json(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION2, "set2", tmp_path)
    shutil.rmtree(os.path.join(tmp_path, "set2", "subfolder2"), ignore_errors=True)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-vv", "--log-format", "json"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/subfolder2' to compress does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Compressing '<DIR>/set2/document.txt'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder1'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION2)


def test_main_subs_missing_valid(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION3A, "set2", tmp_path, add_defs={"subfolder1": DEFINITION3B})
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v"])
    assert ret==0
    captured = capsys.readouterr()
    assert pname(captured.err) == ""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder2'... done. (<DUR>)
 Processing '<DIR>/set2/subfolder1'...
  Compressing '<DIR>/set2/subfolder1/subfolder11'... done. (<DUR>)
  Compressing '<DIR>/set2/subfolder1/subfolder12'... done. (<DUR>)
Completed after <DUR>
"""
    check_generated(tmp_path, actions, dst, "csv")
    check_def(tmp_path, DEFINITION3A)
    
    
def test_main_subs_missing_sub_default(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION3A, "set2", tmp_path, add_defs={"subfolder1": DEFINITION3B})
    shutil.rmtree(os.path.join(tmp_path, "set2", "subfolder1"), ignore_errors=True)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/subfolder1' to recurse into does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder2'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION3A)
    
def test_main_subs_missing_sub_none(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION3A, "set2", tmp_path, add_defs={"subfolder1": DEFINITION3B})
    shutil.rmtree(os.path.join(tmp_path, "set2", "subfolder1"), ignore_errors=True)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v", "--log-off"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/subfolder1' to recurse into does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder2'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION3A)
    
def test_main_subs_missing_sub_csv(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION3A, "set2", tmp_path, add_defs={"subfolder1": DEFINITION3B})
    shutil.rmtree(os.path.join(tmp_path, "set2", "subfolder1"), ignore_errors=True)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v", "--log-format", "csv"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/subfolder1' to recurse into does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder2'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION3A)
    
def test_main_subs_missing_sub_json(capsys, tmp_path):
    actions, dstroot, dst = prepare(DEFINITION3A, "set2", tmp_path, add_defs={"subfolder1": DEFINITION3B})
    shutil.rmtree(os.path.join(tmp_path, "set2", "subfolder1"), ignore_errors=True)
    os.chdir(tmp_path) # !!!
    ret = grebakker.main(["backup", dstroot, str(tmp_path / "set2"), "-v", "-v", "--log-format", "json"])
    assert ret==2
    captured = capsys.readouterr()
    assert pdirtime(captured.err, tmp_path) == """grebakker: error: file/folder '<DIR>/set2/subfolder1' to recurse into does not exist.
"""
    assert pdirtime(captured.out, tmp_path) == """Starting...
Processing '<DIR>/set2'...
 Copying '<DIR>/set2/document.txt'... done. (<DUR>)
 Copying '<DIR>/set2/something_else.csv'... done. (<DUR>)
 Compressing '<DIR>/set2/subfolder2'... done. (<DUR>)
"""
    #check_generated(tmp_path, actions, dst, "csv")
    #check_def(tmp_path, DEFINITION3A)
    
    